
// \file GLView.qt.cpp
// --------------------------------------------------------------------------
// \version $LastChangedRevision$
// \date    $LastChangedDate$
// \author  $LastChangedBy$
// --------------------------------------------------------------------------
// Project: Graphik-Praktikum - Rotations-Geometrie
// --------------------------------------------------------------------------
// (C) Copyright, 2006
// Computer Graphics Group, University of Siegen, Germany
// All Rights Reserved.
//
// This is unpublished proprietary source code.
// The contents of this file may not be disclosed to third parties,
// copied or duplicated in any form, in whole or in part, without the
// prior written permission of the copyright holder.
// --------------------------------------------------------------------------

#include "GL/glew.h"
#include "GLView.qt.h"

char     g_mouseButton = 0;
int      g_oldX = 0, g_oldY = 0;

// defines
#define LBUTTON 1
#define MBUTTON 2
#define RBUTTON 4

const float GLView::s_bgrColor[3] = {0.0f, 0.0f, 0.5f};

/**
 * Constructor
 */
GLView::GLView(QWidget *parent) 
 : QGLWidget(parent)
{
  // create objects
  m_camera = new Camera(0.0, 0.0, 6.0);
  m_camera->setLookAt(1, 0);

  m_MaxCameraDistance = 50.f;
}

/**
 * Destructor 
 */
GLView::~GLView()
{
  delete m_camera;
}

/**
 * Handles new input events when a mouse button is pressed.
 *
 * @param m   mouse event object
 */
void GLView::mousePressEvent(QMouseEvent *m) 
{
  m_lastX = m->x();
  m_lastY = m->y();
}


/**
 * Handles new input events when the mouse is moved.
 *
 * @param m   mouse event object
 */
void GLView::mouseMoveEvent(QMouseEvent *m)
{
  // get delta movement to last position
  int deltaX = m->x() - m_lastX;
  int deltaY = m->y() - m_lastY;

  // check if there has been a movement at all
  if ((deltaX == 0) && (deltaY == 0))
    return;

  // get the buttons involved
  MouseButtons state = m->buttons();
    
  bool lb = (state & LeftButton)  == LeftButton;
  bool mb = (state & MidButton)   == MidButton;
  bool rb = (state & RightButton) == RightButton;

  // process action
  if (lb) {
    m_camera->setPolar(m_camera->polar() + deltaY);
    m_camera->setAzimut(m_camera->azimut() - deltaX);
  }
  else if (rb) {
      m_camera->setDistance( min( (float) m_MaxCameraDistance, (float) (m_camera->distance() - 0.1*deltaY)));
  }
  
  // update state
  m_lastX = m->x();
  m_lastY = m->y();

  updateGL();
}


/**
 * Initialize OpenGL context.
 */ 
void GLView::initializeGL()
{
#ifdef _DEBUG
  glClearColor(0.2f, 0.2f, 0.2f, 1.0f);
#else
  glClearColor(0.0, 0.0, 0.0, 1.0);
#endif
  
	GLenum err = glewInit();
	if (GLEW_OK != err){
		/* Problem: glewInit failed, something is seriously wrong. */
		std::cout << stderr << "Error: %s\n" << glewGetErrorString(err) << std::endl;
	}

  glShadeModel(GL_SMOOTH);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_CULL_FACE);
  glEnable(GL_BLEND);

  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  QString prefix = "../";
#ifndef WIN32
	prefix.append("../");
#endif
	m_Mesh = new ObjLoader( QString("%1resources/bunny.obj").arg(prefix));

  m_camera->setDistance(m_MaxCameraDistance);
}

/**
 * Handle window resize event.
 *
 * @param width    new window width
 * @param height   new window height
 */ 
void GLView::resizeGL(int width, int height)
{
  // update GL state
  glViewport(0, 0, (GLsizei) width, (GLsizei) height);

  m_width = width;
  m_height = height;
  setFrustum();
}

void GLView::setFrustum()
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	float aspect =  (float) m_width/m_height;
	float fovY = 45;
	float zNear = 0.01f;
	float zFar =  3.0*m_MaxCameraDistance;

	GLdouble fW, fH;
	fH = tan( fovY / 360 * M_PI ) * zNear;
	fW = fH * aspect;
	glFrustum( -fW, fW, -fH, fH, zNear, zFar );
	glMatrixMode(GL_MODELVIEW);
}


/**
 * OpenGL draw routine.
 */ 
void GLView::paintGL()
{
	// Clear the buffers
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// Do the Camera transform
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	m_camera->setup();

	// draw the mesh
	glColor3f(1.f,1.f,1.f);
	m_Mesh->draw();
}