#include <QFile>
#include <QTextStream>
#include <QList>
#include <QStringList>
#include "ObjLoader.qt.h"


using namespace std;


ObjLoader::ObjLoader(QObject* parent) 
:	QObject(parent),
	m_Valid(false),
	m_TexID(0)
{
}

ObjLoader::ObjLoader(QString objFilename, QObject* parent)
:	QObject(parent),
	m_Valid(false),
	m_TexID(0)
{
	loadModel(objFilename);
}

ObjLoader::~ObjLoader()
{
	clearModel();
}

void ObjLoader::loadModel(QString filename)
{
	QFile objFile(filename);
	if(!objFile.open(QIODevice::ReadOnly | QIODevice::Text))
		return;

	QTextStream objStream(&objFile);
	QString objLine, prefix;
	
	while(!objStream.atEnd())
	{
		objLine = objStream.readLine();
		QStringList tokens = objLine.split(tr(" "),QString::SkipEmptyParts);
		
		if(tokens.size() == 0)
		{
			continue;
		}

		if(tokens[0].compare(tr("v")) == 0)
		{
			QStringList &vertexCoords = tokens;
			Vertex* v = new Vertex;

			v->x = vertexCoords[1].toFloat();
			v->y = vertexCoords[2].toFloat();
			v->z = vertexCoords[3].toFloat();

			if(vertexCoords.size() == 4)
			{
				v->w = 1.0f;
			}
			else
			{
				v->w = vertexCoords[4].toFloat();
			}
			m_Model.Vertices.append(v);
		}

		if(tokens[0].compare(tr("vt")) == 0)
		{
			QStringList &texCoords = tokens;
			TexCoord* t = new TexCoord;

			t->u = texCoords[1].toFloat();

			if(texCoords.size() == 2)
			{
				t->v = 0.0f;
				t->w = 0.0f;
			}
			else if(texCoords.size() == 3)
			{
				t->v = texCoords[2].toFloat();
				t->w = 0.0f;
			}
			else
			{
				t->v = texCoords[2].toFloat();
				t->w = texCoords[3].toFloat();
			}
			m_Model.TexCoords.append(t);
		}

		if(tokens[0].compare(tr("vn")) == 0)
		{
			QStringList &normals = tokens;
			Normal* n = new Normal;

			n->x = normals[1].toFloat();
			n->y = normals[2].toFloat();
			n->z = normals[3].toFloat();
			
			m_Model.Normals.append(n);
		}

		if(tokens[0].compare(tr("f")) == 0)
		{
			QStringList &faces = tokens;
			Face* f = new Face();
			for(int i = 1; i < faces.size(); ++i)
			{
				QStringList paramList =  faces[i].split(tr("/"));
				switch(paramList.size())
				{
				case 1:
					{
						f->v.append(faces[i].toInt());
						break;
					}
				case 2:
					{
						f->v.append(paramList[0].toInt());
						f->t.append(paramList[1].toInt());
						break;
					}
				case 3:
					{
						f->v.append(paramList[0].toInt());
						if(paramList[1].compare(tr("")) != 0)
							f->t.append(paramList[1].toInt());
						f->n.append(paramList[2].toInt());
						break;
					}
				}
			}
			m_Model.Faces.append(f);
		}
	}
	if(m_Model.Normals.size() == 0)
	{
		generateFaceNormals();
		generateVertexNormals();
	}

	m_Valid = true;

  computeAABB();
}

void ObjLoader::clearModel()
{
	for(int i = 0; i < m_Model.Vertices.size(); ++i)
	{
		delete m_Model.Vertices[i];
	}
	m_Model.Vertices.clear();

	for(int i = 0; i < m_Model.Normals.size(); ++i)
	{
		delete m_Model.Normals[i];
	}
	m_Model.Normals.clear();

	for(int i = 0; i < m_Model.TexCoords.size(); ++i)
	{
		delete m_Model.TexCoords[i];
	}
	m_Model.TexCoords.clear();

	for(int i = 0; i < m_Model.Faces.size(); ++i)
	{
		delete m_Model.Faces[i];
	}
	m_Model.Faces.clear();
	m_Model.VertexToFace.clear();

	m_Valid = false;
}

void ObjLoader::draw(bool bDrawAABB, bool bCentered)
{
  glPushMatrix();
    if(bCentered) { 
      glTranslatef(-0.5f*(m_MaxAABB.x-m_MinAABB.x)-m_MinAABB.x,
                   -0.5f*(m_MaxAABB.y-m_MinAABB.y)-m_MinAABB.y,
                   -0.5f*(m_MaxAABB.z-m_MinAABB.z)-m_MinAABB.z);
    }

	  if (m_TexID != 0) {
		  glEnable(GL_TEXTURE_2D);
		  glBindTexture(GL_TEXTURE_2D, m_TexID);
	  }
	  QVector<Vertex*>& v = m_Model.Vertices;
	  QVector<Normal*>& n = m_Model.Normals;
	  QVector<TexCoord*>& t = m_Model.TexCoords;
	  for(int i = 0; i < m_Model.Faces.size(); ++i)
	  {
		  glBegin(GL_POLYGON);
		  Face* f = m_Model.Faces[i];
		  for(int j = 0; j < f->v.size(); ++j)
		  {
			  /*if(f->faceNormal)
			  {
				  glNormal3f(f->faceNormal->x,f->faceNormal->y,f->faceNormal->z);
			  }*/
			  if(f->n.size() > 0)
			  {
				  glNormal3f(n[f->n[j]-1]->x, n[f->n[j]-1]->y, n[f->n[j]-1]->z);
			  }

			  if(f->t.size() > 0)
			  {
				  glTexCoord3f(t[f->t[j]-1]->u, t[f->t[j]-1]->v, t[f->t[j]-1]->w);
				  //glColor3f(t[f->t[j]-1]->u, t[f->t[j]-1]->v, t[f->t[j]-1]->u * t[f->t[j]-1]->v);
			  }

			  glVertex4f(v[f->v[j]-1]->x, v[f->v[j]-1]->y, v[f->v[j]-1]->z, v[f->v[j]-1]->w);
		  }
		  glEnd();
	  }

    if(bDrawAABB) { 
      drawAABB();
    }
  glPopMatrix();
}

void ObjLoader::generateFaceNormals()
{
	QVector<Face*> &vec = m_Model.Faces;

	for(int i = 0; i < vec.size(); ++i)
	{
		for(int j = 0; j < vec[i]->v.size(); ++j)
		{
			Vertex* v = m_Model.Vertices[vec[i]->v[j]-1];
			vec[i]->n.append(vec[i]->v[j]);
			m_Model.VertexToFace.insertMulti(v,vec[i]);
		}
		Vertex v1, v2, v3;
		v1 = *m_Model.Vertices[vec[i]->v[0]-1];
		v2 = *m_Model.Vertices[vec[i]->v[1]-1];
		v3 = *m_Model.Vertices[vec[i]->v[2]-1];
		vec[i]->faceNormal = new Normal;
		*vec[i]->faceNormal = (v3-v2)%(v1-v2);
		vec[i]->faceNormal->normalize();
	}
}

void ObjLoader::generateVertexNormals()
{
	QMap<Vertex*,Face*> &m = m_Model.VertexToFace;
	QVector<Vertex*> &v = m_Model.Vertices;

	for(int i = 0; i < v.size(); ++i)
	{
		Normal* avgNormal = new Normal;
		QList<Face*> l = m.values(v[i]);
		QListIterator<Face*> it(l);

		while(it.hasNext())
		{
			Face* f = it.next();
			*avgNormal = *avgNormal + *f->faceNormal;
		}

		avgNormal->normalize();
		m_Model.Normals.append(avgNormal);
	}
}

bool ObjLoader::updateTexture(QString texFilename)
{
	QImage t;
	QImage b;

	if (!b.load(texFilename)){
		//std::cout << "texture: failed to load image: " << file << std::endl;
		return false;
	}
	t = QGLWidget::convertToGLFormat( b );

	if (m_TexID == 0)
		glGenTextures(1, &m_TexID);

	//create a new texture object and assign a texture name to it
	glBindTexture(GL_TEXTURE_2D, m_TexID);

	//glTexParameteri(GLenum target, GLenum pname, TYPE param);
	//sets various parameters that control how a texture is stored
	//in a texture object;
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_TEXTURE_WRAP_S);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_TEXTURE_WRAP_T);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);


	glTexImage2D(
		GL_TEXTURE_2D,                //target
		0,					                  //level
		GL_RGBA,				              //internalFormat
		t.width(),                    //width
		t.height(),                   //height
		0,					                  //boarder
		GL_RGBA,				              //format
		GL_UNSIGNED_BYTE,           	//type
		t.bits()                      // texels
		);

	return true;
}

void ObjLoader::computeAABB()
{
  if(!m_Valid) return;

  m_MinAABB.x = m_MinAABB.y = m_MinAABB.z = FLT_MAX;
  m_MaxAABB.x = m_MaxAABB.y = m_MaxAABB.z = FLT_MIN;

  for(int i=0; i<m_Model.Vertices.size(); ++i)
  {
    m_MinAABB.x = min(m_MinAABB.x, m_Model.Vertices[i]->x);
    m_MinAABB.y = min(m_MinAABB.y, m_Model.Vertices[i]->y);
    m_MinAABB.z = min(m_MinAABB.z, m_Model.Vertices[i]->z);

    m_MaxAABB.x = max(m_MaxAABB.x, m_Model.Vertices[i]->x);
    m_MaxAABB.y = max(m_MaxAABB.y, m_Model.Vertices[i]->y);
    m_MaxAABB.z = max(m_MaxAABB.z, m_Model.Vertices[i]->z);
  }
}


void ObjLoader::drawAABB()
{
  if(!m_Valid) return;

  glBegin(GL_LINE_LOOP);
    glVertex3f(m_MinAABB.x, m_MinAABB.y, m_MinAABB.z);
    glVertex3f(m_MinAABB.x, m_MaxAABB.y, m_MinAABB.z);
    glVertex3f(m_MaxAABB.x, m_MaxAABB.y, m_MinAABB.z);
    glVertex3f(m_MaxAABB.x, m_MinAABB.y, m_MinAABB.z);
	glEnd();

  glBegin(GL_LINES);
    glVertex3f(m_MinAABB.x, m_MinAABB.y, m_MaxAABB.z);
    glVertex3f(m_MinAABB.x, m_MinAABB.y, m_MinAABB.z);
    glVertex3f(m_MaxAABB.x, m_MaxAABB.y, m_MaxAABB.z);
    glVertex3f(m_MaxAABB.x, m_MaxAABB.y, m_MinAABB.z);
    glVertex3f(m_MinAABB.x, m_MaxAABB.y, m_MaxAABB.z);
    glVertex3f(m_MinAABB.x, m_MaxAABB.y, m_MinAABB.z);
    glVertex3f(m_MaxAABB.x, m_MinAABB.y, m_MaxAABB.z);
    glVertex3f(m_MaxAABB.x, m_MinAABB.y, m_MinAABB.z);
	glEnd();

  glBegin(GL_LINE_LOOP);
    glVertex3f(m_MinAABB.x, m_MinAABB.y, m_MaxAABB.z);
    glVertex3f(m_MinAABB.x, m_MaxAABB.y, m_MaxAABB.z);
    glVertex3f(m_MaxAABB.x, m_MaxAABB.y, m_MaxAABB.z);
    glVertex3f(m_MaxAABB.x, m_MinAABB.y, m_MaxAABB.z);
	glEnd();



}