#pragma once

#include <QObject>
#include <QVector>
#include <QMap>
#include <QtOpenGL/QtOpenGL>
#include <float.h>

class ObjLoader : public QObject
{
	Q_OBJECT
public:
	struct Normal
	{
		float x;
		float y;
		float z;

		Normal operator-(const Normal& n) const
		{
			Normal out;
			out.x = x - n.x;
			out.y = y -  n.y;
			out.z = z -  n.z;
			return out;
		}

		Normal operator+(const Normal& n) const
		{
			Normal out;
			out.x = x + n.x;
			out.y = y +  n.y;
			out.z = z +  n.z;
			return out;
		}

		Normal operator%(const Normal& n) const
		{
			Normal out;
			out.x = y*n.z - z * n.y;
			out.y = z*n.x - x * n.z;
			out.z = x*n.y - y * n.x;
			return out;
		}

		Normal operator/(const float f) const
		{
			Normal out;
			out.x = x / f;
			out.y = y / f;
			out.z = z / f;
			return out;
		}
		Normal()
		{
			x = y = z = 0.f;
		}

		void normalize()
		{
			float one_over_length = 1/sqrtf(x*x + y*y + z*z);
			x *= one_over_length;
			y *= one_over_length;
			z *= one_over_length;

		}
	};

	struct Vertex
	{
		float x;
		float y;
		float z;
		float w;
		Normal operator-(const Vertex& v) const
		{
			Normal n;
			n.x = x/w - v.x/v.w;
			n.y = y/w - v.y/v.w;
			n.z = z/w - v.z/v.w;
			return n;
		}
		Vertex operator*(const float& f) const
		{
			Vertex v;
			v.x = x * f;
			v.y = y * f;
			v.z = z * f;
			return v;
		}

		float distance(const Vertex& v)
		{
			float x1 = x - v.x;
			float y1 = y - v.y;
			float z1 = z - v.z;
			return sqrtf(x1*x1 + y1*y1 + z1*z1);
		}
	};

	struct TexCoord
	{
		float u;
		float v;
		float w;
	};

	struct Face
	{
		QVector<int> v;
		QVector<int> n;
		QVector<int> t;
		Normal* faceNormal;
		Face() : faceNormal(0){}
	};

	struct Model
	{
		QVector<Vertex*> Vertices;
		QVector<Normal*> Normals;
		QVector<TexCoord*> TexCoords;

		QVector<Face*> Faces;
		QMap<Vertex*,Face*> VertexToFace;
	};

	ObjLoader(QObject* parent = 0);
	ObjLoader(QString objFilename, QObject* parent = 0);
	~ObjLoader();

	void loadModel(QString objFilename);
	void clearModel();
	bool isValid(){return m_Valid;}
	void draw(bool bDrawAABB=false, bool bCentered=true);
	bool updateTexture(QString texFilename);
	Vertex getMinAABB(){return m_MinAABB;}
	Vertex getMaxAABB(){return m_MaxAABB;}
	void drawAABB();

  protected:
	void generateFaceNormals();
	void generateVertexNormals();
	Vertex m_MinAABB;
	Vertex m_MaxAABB;
	Model m_Model;
	bool m_Valid;
	GLuint m_TexID;

private:

  void computeAABB();

};